/************************************************************************

 *    TABU SEARCH FOR THE CLASSICAL GENERALIZED ASSIGNMENT PROBLEM      *

 *                                                                      *

 *    Features:   1) Single level agents                                *

 *                2) Ejection chains                                    *

 *                3) Multiple dynamic tabu lists with long term memory  *

 *                4) Strategic Oscillation                              *

 *                                                                      *

 *              Copyright (C) 1991  by  Manuel Laguna                   *

 *                                      Jose Luis Gonzalez Velarde      *

 *                                      James P. Kelly                  *

 *                                      Fred Glover                     *

 *                                                                      *

 *    Last revision: July 15, 1992.                      Version: 4.3   *

 ************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <time.h>           /* to use clock() and randomize()           */

#define BIG_NUM 999999
#define MAX(X,Y)    ( (X) > (Y) ? (X) : (Y) )
#define MIN(X,Y)    ( (X) < (Y) ? (X) : (Y) )
#define ABS(X)      ( (X) < 0 ? -(X) : (X) )

struct move_info
{
	int task1;

	int task2;

	int agent1;

	int agent2;

	long int value;

	long int inf;

};



struct sol_info
{
	long int value;

	long int inf;

};



/************************************************************************

 *                       FUNCTION PROTOTYPES                            *

 ************************************************************************/



void tabu_search(char *name, int max_time, int m, int n, int **c, int **a,
int *b);

void input_data(FILE *f1, int m, int n, int **c, int **a, int *b);

void initialization(int m, int n, int **c, int **a, int *b, int *curr_asgn,
struct sol_info *current, int *eject, int **best_agent, int **task_list,

long int *b_prime, int **tabu_time, int **agent_rank, int **freq);

void indexx(int n, int *arrin, int *indx);

void update_best(int n, int *curr_asgn, struct sol_info current, int *best_asgn, struct sol_info *best);

void best_move(int n, int **c, int **a, int *curr_asgn, long int *b_prime,
int **task_list, int *eject, int **tabu_time, int iter, int **best_agent,

struct sol_info current, struct sol_info best, struct move_info *move,
int flag, long int thresh);

void check_criteria(int task1, int agent1, int task2, int agent2, int agent0, struct move_info *besti, struct move_info *bestv, int **tabu_time,

long int *b_prime, struct sol_info current, struct sol_info best, int iter,
int **c, int **a, long int b_dbl_prime);

void update_all(int iter, int m, int n, struct move_info move, int *curr_asgn,

long int *b_prime, int *eject, int **best_agent, int **task_list, int **agent_rank, int **tabu_time, int **a, int **freq);

int new_eject(int task, int agent, long int *b_prime, int **best_agent,
int **a);

void delete_task(int task, int agent, int **task_list);

void insert_task(int task, int agent, int **task_list);

void tabu_update(int iter, int task, int agent0, int agent1, int k, int **rank,
int **tabu_time, int **freq, int max_freq, int m);

void execute_move(struct move_info move, int *curr_asgn, struct sol_info *current);



/***************** Memory Allocation Prototypes *************************/



int *ivector(int nl, int nh);

void free_ivector(int *v, int nl);

long int *livector(int nl, int nh);

void free_livector(long int *v, int nl);

int **imatrix(int nrl, int nrh, int ncl, int nch);

void free_imatrix(int **m, int nrl, int nrh, int ncl);

void nrerror(char *error_text);



/************************************************************************

 *                         MAIN FUNCTION                                *

 ************************************************************************/



void main(int argc, char **argv)

{
	int n;                  /* number of tasks                          */

	int m;                  /* number of agents                         */

	int **c;                /* cost of assigning task j to agent i      */

	int **a;                /* resources used by agent i on task j      */

	int *b;                 /* available resources for agent i          */

	int s;



	FILE *f1;               /* pointer to input file                    */



	if (argc != 3) nrerror("TSGAP: < file name > < cpu seconds >");
	f1 = fopen(argv[2],"r");
	printf("\nArchivo a leer: %s  \n", argv[2]);
	if (f1 == NULL) nrerror("Unable to open data file...");

	fscanf(f1,"%d %d", &m, &n);

	c = imatrix(1, m, 1, n);

	a = imatrix(1, m, 1, n);

	b = ivector(1, m);

	input_data(f1, m, n, c, a, b);

	fclose(f1);

        s = atoi(argv[1]);

	tabu_search(argv[2], s, m, n, c, a, b);

}



/************************************************************************

 *                           TABU SEARCH                                *

 ************************************************************************/



void tabu_search(char *name, int max_time, int m, int n, int **c, int **a,
int *b)

{
      int j;                     /* loop index                            */

	int iter;                  /* iteration counter                     */

	int iter_best;             /* iteration where best was found        */

	int thresh_count;          /* no. of iterations with same thresh    */

	int *curr_asgn;            /* current assignment                    */

	int *best_asgn;            /* best assignment                       */

	int *eject;                /* best agent where to "eject" task j    */

	int **agent_rank;          /* ranking of of each best agent         */

	int **best_agent;          /* index of ith best agent for task j    */

	int **task_list;           /* index of tasks assigned to agent i    */

	int **tabu_time;           /* iter when j can be assigned to i      */

	int **freq;                /* move frequencies                      */

	

	long int thresh;           /* infesibility threshold                */

	long int current_best;     /* best objective from last time in      */

	long int *b_prime;         /* slack on agent i                      */

	

	struct sol_info current;   /* current solution                      */

	struct sol_info best;      /* best solution                         */

	struct move_info move;     /* information for best move             */



	FILE *f1;                  /* pointer to input file                 */

	

	clock_t start_t;           /* starting time                         */

	clock_t best_t;            /* time to best                          */



	curr_asgn = ivector(1, n);

	best_asgn = ivector(1, n);

	eject = ivector(1, n);
	agent_rank = imatrix(1, n, 1, m);
	best_agent = imatrix(1, n, 0, m+1);
	task_list = imatrix(1, m, 0, n);
	tabu_time = imatrix(1, m, 1, n);

	freq = imatrix(1, n, 1, m);

	b_prime = livector(1, m);

	

	start_t = clock();

	initialization(m, n, c, a, b, curr_asgn, &current, eject, best_agent, task_list, b_prime, tabu_time, agent_rank, freq);

	update_best(n, curr_asgn, current, best_asgn, &best);

	iter = 0;

	iter_best = 0;

	thresh_count = 0;

	thresh = BIG_NUM;

	current_best = - BIG_NUM;


	do 
	{
	++iter;

		best_move(n, c, a, curr_asgn, b_prime, task_list, eject, tabu_time, iter, best_agent, current, best, &move, 0, thresh);

		/*printf("move.value=%lf\n", move.value);
*/
		if (move.value == -BIG_NUM)

			best_move(n, c, a, curr_asgn, b_prime, task_list, eject,
tabu_time, iter, best_agent, current, best, &move, 1, thresh);

		if (!current.inf) 
		{
	if (move.value > current_best) 
		        {
 current_best = move.value;

			   thresh_count = 0;

			} 
			else
			   ++thresh_count;

			if (thresh_count >= m*n) thresh = 0;

		}

		if (current.inf && !move.inf)
		{
	current_best = move.value;

			thresh = move.value;

			thresh_count = 0;

		} 

		update_all(iter, m, n, move, curr_asgn, b_prime, eject, best_agent, task_list, agent_rank, tabu_time, a, freq);		

		execute_move(move, curr_asgn, &current);

		/*printf(" current.inf=(%lf)  best.inf=(%lf)   best.value=(%lf)\n",current.inf, best.inf, best.value); 
*/
		if (current.inf < best.inf || (current.inf == best.inf && current.value > best.value))
		{
	update_best(n, curr_asgn, current, best_asgn, &best);

			iter_best = iter;

			best_t = clock();

		}

	} while ((clock()-start_t)/(float)CLOCKS_PER_SEC < max_time);

	
	printf("\n*** File name: %s  Total Iterations = %5d   CPU Time = %10.2f\n", name, iter, (clock()-start_t)/(float)CLOCKS_PER_SEC);

	printf("*** Best Found = (%4li, %4li)  Iter = %5d   CPU Time = %10.2f\n",
best.value, best.inf, iter_best, (best_t-start_t)/(float)CLOCKS_PER_SEC);

        printf("\n    JOB        AGENT\n");

        for (j = 1; j <= n; ++j)
printf("%6d      %6d\n", j, best_asgn[j]);

}



/************************************************************************

 *                         INITIALIZATION                               *

 ************************************************************************/

void initialization(int m, int n, int **c, int **a, int *b, int *curr_asgn,

struct sol_info *current, int *eject, int **best_agent, int **task_list,
long int *b_prime, int **tabu_time, int **agent_rank, int **freq)

{
	int i;                  /* agent index                              */

	int j;                  /* task  index                              */

	int k;                  /* number of entries in 'cost' vector       */

	int *cost;              /* temporary cost vector (sorting key)      */

	int *indxc;             /* cost index (ascending order)             */

	int *indxr;             /* ratio index (ascending order)            */

	int *ratio;             /* knapsack ratio                           */

	

	cost = ivector(1, m);

	indxc = ivector(1, m);

	indxr = ivector(1, m);

	ratio = ivector(1, m);



	current->value = 0L;

	for (i = 1; i <= m; ++i)
	{
	eject[i] = 0;

		task_list[i][0] = 0;

		b_prime[i] = (long) b[i];

	}

	for (j = 1; j <= n; ++j) 
	{
 	for (i = 1; i <= m; ++i)
		{
	cost[i] = c[i][j];

			ratio[i] = (a[i][j] > 0) ? (int)((cost[i]*1000)/(float)a[i][j]) : -1;

			best_agent[j][i] = 0;

			tabu_time[i][j] = 0;

			agent_rank[j][i] = m+1;

			freq[j][i] = 0;

		}

		best_agent[j][m+1] = 0;

		indexx(m, cost, indxc);

		indexx(m, ratio, indxr);

		k = 0;

		best_agent[j][0] = 0;

		for (i = 1; i <= m; ++i) 
		{
	if (c[indxc[i]][j] >= 0) 
			{
	best_agent[j][++k] = indxc[i];

				++best_agent[j][0];

			}

		}

		for (i = 1; i <= m; ++i)

			if (c[indxr[i]][j] >= 0)

				agent_rank[j][indxr[i]] = i - m + best_agent[j][0];

		curr_asgn[j] = (best_agent[j][0]) ? best_agent[j][1] : 0;

		if (curr_asgn[j]) 
		{
	current->value += c[curr_asgn[j]][j];

			b_prime[curr_asgn[j]] -= a[curr_asgn[j]][j];

			task_list[curr_asgn[j]][++task_list[curr_asgn[j]][0]] = j;

		}

	}

	current->inf = 0L;

	for (i = 1; i<= m; ++i)

		if (b_prime[i] < 0) current->inf -= b_prime[i];

	for (j = 1; j <= n; ++j)

		eject[j] = new_eject(j, curr_asgn[j], b_prime, best_agent, a);

	free_ivector(cost, 1);

	free_ivector(indxc, 1);

	free_ivector(indxr, 1);

	free_ivector(ratio, 1);

}

		

/************************************************************************

 *                  CONSTRUCTION OF AN INDEX TABLE                      *

 ************************************************************************/

void indexx(int n, int *arrin, int *indx)

{
 int i, j ;

   int t;

   int **a;

   a = imatrix(1,n,1,2);

   for(i=1; i<=n; ++i) 
   {
  a[i][1] = arrin[i];

       a[i][2] = i;

   }

   i = 1;

   while(i<n) 
   {
 j = n;

      while(j>i) 
      {
  if (a[j][1] > a[j-1][1]) 
          {
  t = a[j][1];

	      a[j][1] = a[j-1][1];

	      a[j-1][1] = t;

	      t = a[j][2];

	      a[j][2] = a[j-1][2];

	      a[j-1][2] = t;

	  }

	  j = j-1;

      }

      i++;

   }

   for(i=1; i<=n; ++i)

   indx[i] = a[i][2];



   free_imatrix(a,1,n,1);

}

		

/************************************************************************

 *                      UPDATE BEST SOLUTION                            *

 ************************************************************************/



void update_best(int n, int *curr_asgn, struct sol_info current,
int *best_asgn, struct sol_info *best)

{
	int j;                  /* task  index                              */

	for (j = 1; j <= n; ++j)

		best_asgn[j] = curr_asgn[j];

	*best = current;

}



/************************************************************************

 *                           BEST MOVE                                  *

 ************************************************************************/



void best_move(int n, int **c, int **a, int *curr_asgn, long int *b_prime,

int **task_list, int *eject, int **tabu_time, int iter, int **best_agent,
struct sol_info current, struct sol_info best, struct move_info *move,
int flag, long int thresh)

{
	int i0;                 /* agent 0 index                            */

	int i1;                 /* agent 1 index                            */

	int i2;                 /* agent 2 index                            */

	int j1;                 /* task  1 index                            */

	int j2;                 /* task  2 index                            */

	int k;                  /* task index for agent i                   */

	int pnt;                /* pointer to best_agent list               */

	long int b_dbl_prime;   /* updated b_prime for agent 1              */

	int static way_out = 0; /* if = 1 then moving out of feasible R     */

	

	struct move_info bestv;

	struct move_info besti;

	

	bestv.value = - BIG_NUM;

	besti.value = - BIG_NUM;

	bestv.inf = BIG_NUM;

	besti.inf = BIG_NUM;

	for (j1 = 1; j1 <= n; ++j1) 
	{
pnt = 0;

	  i0 = curr_asgn[j1];

	  do 
	  {
i1 = best_agent[j1][++pnt];

	    if (i1 && i1 != i0) 
	    {
b_dbl_prime = b_prime[i1] - a[i1][j1];

	      if (b_dbl_prime >= 0)

	        check_criteria(j1, i1, 0, 0, i0, &besti, &bestv, tabu_time, b_prime, current, best, iter, c, a, b_dbl_prime);

	      else 
	      { for (k = 1; k <= task_list[i1][0]; ++k) 
	        {
j2 = task_list[i1][k];

	          i2 = eject[j2];

	          b_dbl_prime = b_prime[i1]+(long)a[i1][j2]-(long)a[i1][j1];

	          if (i2 && (flag ||  b_dbl_prime >= 0)) 
	          {
check_criteria(j1, i1, j2, i2, i0, &besti, &bestv, tabu_time, b_prime, current, best, iter, c, a, b_dbl_prime);

	            if (i2 != i0 && c[i0][j2] >= 0)

 	                check_criteria(j1, i1, j2, i0, i0, &besti, &bestv, tabu_time, b_prime, current, best, iter, c, a, b_dbl_prime);

	          }

	        }

	      } 

	    }

	  } while (i1);

	}

	if (current.inf) 
	  {
if (way_out) 
	      {
 if (current.value <= best.value) *move = bestv;

		 else {
 *move = besti;
way_out = 0;
}

	      }
	    else *move = besti;
	  }
	else
	  {
 if (!besti.inf && besti.value > thresh)
*move = besti;

	     else {
*move = bestv;
way_out = 1;
}

	  }

}	



/************************************************************************

 *                       CHECKING CRITERIA                              *

 ************************************************************************/

void check_criteria(int task1, int agent1, int task2, int agent2, int agent0, struct move_info *besti, struct move_info *bestv, int **tabu_time,

long int *b_prime, struct sol_info current, struct sol_info best, int iter,
int **c, int **a, long int b_dbl_prime)

{
	int tabu_status;             /* if > 0 move is tabu                 */

	long int inf;                /* infeasibility                       */

	long int value;              /* objective function value            */

	long int bp;                 /* temporary storage for b_prime       */

	

	inf = (b_prime[agent0] < 0) ? MIN(-b_prime[agent0], a[agent0][task1]) : 0;

	inf = current.inf - inf;

	value = current.value - c[agent0][task1] + c[agent1][task1];

	if (task2) 
	{
	value += -c[agent1][task2] + c[agent2][task2];

		if (b_dbl_prime >= 0)

			inf += MIN(b_prime[agent1], 0);

		else

			inf += (b_prime[agent1] >= 0) ? -b_dbl_prime :

			b_prime[agent1] - b_dbl_prime;

		bp = (agent2 != agent0) ? b_prime[agent2] : b_prime[agent0]+a[agent0][task1];

		inf += (bp > 0) ? -MIN(bp - a[agent2][task2], 0) : a[agent2][task2];

	}

	tabu_status = 0;

	if (tabu_time[agent1][task1] >= iter) ++tabu_status;

	if (task2 && tabu_time[agent2][task2] >= iter) ++tabu_status;

	if (!tabu_status || (task2 && tabu_status < 2) ||
(value > best.value && inf <= best.inf)) 
	{
	if (value > bestv->value || (value == bestv->value && inf < bestv->inf)) 
		{
	bestv->task1 = task1;

			bestv->agent1 = agent1;

			bestv->task2 = task2;

			bestv->agent2 = agent2;

			bestv->value = value;

			bestv->inf = inf;

		}

		if (inf < besti->inf || (value > besti->value && inf == besti->inf)) 
		{
	besti->task1 = task1;

			besti->agent1 = agent1;

			besti->task2 = task2;

			besti->agent2 = agent2;

			besti->value = value;

			besti->inf = inf;

		}

	}

}



/************************************************************************

 *                         UPDATE ALL                                   *

 ************************************************************************/



void update_all(int iter, int m, int n, struct move_info move, int *curr_asgn,

long int *b_prime, int *eject, int **best_agent, int **task_list, int **agent_rank,
 int **tabu_time, int **a, int **freq)

{
	int j;                  /* task  index                              */

	int i;                  /* agent index                              */

	int static max_freq = 1;

	b_prime[curr_asgn[move.task1]] += a[curr_asgn[move.task1]][move.task1];

	b_prime[move.agent1] -= a[move.agent1][move.task1];

	if (move.task2) 
	{
	b_prime[move.agent1] += a[move.agent1][move.task2];

		b_prime[move.agent2] -= a[move.agent2][move.task2];

	}

	for (j = 1; j <= n; ++j) 
	{
	i = curr_asgn[j];

		if (j == move.task1) i = move.agent1;

		if (j == move.task2) i = move.agent2;

		eject[j] = new_eject(j, i, b_prime, best_agent, a);

	}

	delete_task(move.task1, curr_asgn[move.task1], task_list);

	insert_task(move.task1, move.agent1, task_list);

	tabu_update(iter, move.task1, curr_asgn[move.task1], move.agent1,

	best_agent[move.task1][0], agent_rank, tabu_time, freq, max_freq, m);

	++freq[move.task1][move.agent1];

	if (move.task2) 
	{
	delete_task(move.task2, move.agent1, task_list);

		insert_task(move.task2, move.agent2, task_list);

		tabu_update(iter, move.task2, curr_asgn[move.task2], move.agent2,

		best_agent[move.task2][0], agent_rank, tabu_time, freq, max_freq, m);

		++freq[move.task2][move.agent2];

		max_freq = (freq[move.task2][move.agent2] > max_freq) ?

		freq[move.task2][move.agent2] : max_freq;

	}

	max_freq = (freq[move.task1][move.agent1] > max_freq) ?

	freq[move.task1][move.agent1] : max_freq;

}



int new_eject(int task, int agent, long int *b_prime, int **best_agent, int **a)

{
	int i;                  /* agent index                              */

	int eject_agent;        /* index of new eject agent for task        */

	int flag;               /* to indicate a feasible ejection found    */

	long int min_inf;       /* min infeasible found                     */

	long int inf;           /* infeasibility for current chain          */     

	min_inf = BIG_NUM;

	eject_agent = 0;

	flag = 0;

	i = 1;

	do {

		if (best_agent[task][i] != agent) 
		{	inf = (b_prime[best_agent[task][i]] > 0) ?

			a[best_agent[task][i]][task] - b_prime[best_agent[task][i]] : 

			a[best_agent[task][i]][task];

			if (inf <= 0) 
			{
	eject_agent = best_agent[task][i];

				flag = 1;

			} else

				if (inf < min_inf) eject_agent = best_agent[task][i];

		}

		++i;

	} while (best_agent[task][i] && !flag);

	return(eject_agent);

}

			

void delete_task(int task, int agent, int **task_list)

{
	int j;                  /* task  index                              */

	for (j = 1; j <= task_list[agent][0]; ++j) 
	{
	if (task_list[agent][j] == task) 
		{
	task_list[agent][j] = task_list[agent][task_list[agent][0]];

			--task_list[agent][0];

			break;

		}

	}

}



void insert_task(int task, int agent, int **task_list)

{
	++task_list[agent][0];

	task_list[agent][task_list[agent][0]] = task;

}



void tabu_update(int iter, int task, int agent0, int agent1, int k, int **rank,
int **tabu_time, int **freq, int max_freq, int m)

{
	int delta;              /* ranking difference                       */

	float x;                /* independent variable                     */

	delta = rank[task][agent0] - rank[task][agent1];

	x = delta / (float)(k - 1); 

	tabu_time[agent0][task] = iter + (int)(0.5*k*(3.0 + x));

	tabu_time[agent0][task] += (int)(m * freq[task][agent0]/(float)max_freq);

}



/************************************************************************

 *                        EXECUTE MOVE                                  *

 ************************************************************************/



void execute_move(struct move_info move, int *curr_asgn, struct sol_info *current)

{
	curr_asgn[move.task1] = move.agent1;

	if (move.task2) curr_asgn[move.task2] = move.agent2;

	current->value = move.value;

	current->inf = move.inf;

}

		

/************************************************************************

 *                      INPUT DATA FUNCTION                             *

 ************************************************************************/



void input_data(FILE *f1, int m, int n, int **c, int **a, int *b)

{
	int i;                  /* agent index                              */

	int j;                  /* task  index                              */

	int point;

	for (i = 1; i <= m; ++i) 
	   { for (j = 1; j <= n; ++j)

		fscanf(f1, "%d", &c[i][j]);

	     fscanf(f1, "%d", &point);

	   }

	for (i = 1; i <= m; ++i) 
	   { for (j = 1; j <= n; ++j)

		fscanf(f1, "%d", &a[i][j]);

	     fscanf(f1, "%d", &point);

	   }

	for (i = 1; i <= m; ++i)

	   fscanf(f1, "%d", &b[i]);

}



/************************************************************************

 *                  MEMORY ALLOCATION FUNCTIONS                         *

 ************************************************************************/



int *ivector(int nl, int nh)

{
	int *v;

	v = (int *)malloc((unsigned)(nh-nl+1)*sizeof(int));

	if (!v) nrerror("Memory allocation failure in ivector()");

	return v - nl;

}



void free_ivector(int *v, int nl)

{
	free((char*) (v + nl));

}



int **imatrix(int nrl, int nrh, int ncl, int nch)

{
	int i;

	int **m;

	m = (int **)malloc((unsigned)(nrh - nrl +1)*sizeof(int*));

	if (!m) nrerror("allocation failure 1 in imatrix()");

	m -= nrl;

	for(i = nrl; i <= nrh ; i++)
	{
 	m[i]= (int*)malloc((unsigned)(nch - ncl +1)*sizeof(int));

		if (!m[i]) nrerror("allocation failure 2 in imatrix()");

		m[i] -= ncl;

	}

	return m;

}



void free_imatrix(int **m, int nrl, int nrh, int ncl)

{
	int i;

	for (i = nrh; i >= nrl; i--)

	     free((char*) (m[i] + ncl));

	free((char *)(m + nrl));

}



long int *livector(int nl, int nh)

{
	long int *v;

	v = (long int *)malloc((unsigned)(nh-nl+1)*sizeof(long int));

	if (!v) nrerror("allocation failure in livector()");

	return v - nl;

}



void free_livector(long int *v, int nl)

{
	free((char*) (v + nl));

}



void nrerror(char *error_text)

{
	fprintf(stderr,"\n%s\n\n",error_text);
	exit(1);

}

